/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

int verb_install(int argc, char *argv[], void *userdata);
int verb_remove(int argc, char *argv[], void *userdata);
int verb_is_installed(int argc, char *argv[], void *userdata);
